# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

# The MIT License
#
# Copyright (c) 2017 Tammo Ippen, tammo.ippen@posteo.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from os.path import commonprefix

from geohash2 import encode


def bbox(shp):
    '''Compute the bounding box of the given shape (Polygon and MultiPolygon allowed).

    If `shp` already contains a `bbox` member, return that instead.

    Parameters:
        shp: Dict[str, Any]  geojson feature, that is either a Polygon or MultiPolygon

    Returns:
        Tuple[float, float, float, float]: Bounding box of `shp`: minlng, minlat, maxlng, maxlat
    '''
    if shp.get('bbox'):
        return shp['bbox']

    if shp['type'] == 'Polygon':
        coords = (p for p in shp['coordinates'][0])
    elif shp['type'] == 'MultiPolygon':
        coords = (p for poly in shp['coordinates'] for p in poly[0])
    else:
        raise ValueError('Expect Polygon or MultiPolygon.')

    minlng, maxlng, minlat, maxlat = None, None, None, None

    for lng, lat in coords:
        if minlng is None:
            minlng, maxlng = lng, lng
            minlat, maxlat = lat, lat
        minlat = min(lat, minlat)
        minlng = min(lng, minlng)
        maxlat = max(lat, maxlat)
        maxlng = max(lng, maxlng)

    return minlng, minlat, maxlng, maxlat


def in_bbox(p, bbox):
    '''Test, whether point p (lng,lat) is in bbox (minlng, minlat, maxlng, maxlat)

    Parameters:
        p: Tuple[float, float]  2D point (lng, lat) (WGS84) Longitude (-180, 180) Latitude (-90, 90)
        bbox: Tuple[float, float, float, float]  Bounding box, (minlng, minlat, maxlng, maxlat)

    Returns:
        bool: True, if point is in bbox, False otherwise.
    '''
    lng, lat = p
    minlng, minlat, maxlng, maxlat = bbox
    return minlng <= lng <= maxlng and minlat <= lat <= maxlat


def bbox_hash(bbox):
    '''Get geohash from rectangle covering the complete bbox.

    Parameters:
        bbox: Tuple[float, float, float, float]  Bounding box, (minlng, minlat, maxlng, maxlat)

    Returns:
        str: geohash covering the complete bbox.
    '''
    minlng, minlat, maxlng, maxlat = bbox
    ll = encode(latitude=minlat, longitude=minlng)
    ur = encode(latitude=maxlat, longitude=maxlng)

    return commonprefix((ll, ur))


def ccw(a, b, c):
    '''Tests whether the turn formed by a, b, and c is CCW

    Returns:
        > 0 for c left of the line through a and b (ccw)
        = 0 for p2 on the line
        < 0 for p2  right of the line (not ccw)
    '''
    return (b[0] - a[0]) * (c[1] - a[1]) - (c[0] - a[0]) * (b[1] - a[1])


def is_convex(ring):  # TODO
    n = len(ring)
    for i in range(2, n):
        p0 = ring[(i - 2) % n]
        p1 = ring[(i - 1) % n]
        p2 = ring[i % n]
        if ccw(p0, p1, p2) > 0:
            return False
    return True


def winding_number(p, ring):
    '''Return the winding number of `p` wrt `ring`.

    If the winding number is 0, then p is outside of the ring.

    Parameters:
        p: Tuple[float, float]           2D Point.
        ring: List[Tuple[float, float]]  Ring of Points in CCW orientation (ring[0] == ring[-1]).

    Returns:
        int: the winding number (=0 only if `p` is outside `ring`)
    '''
    n = len(ring)
    wn = 0
    for i in range(1, n):
        p1 = ring[i - 1]
        p2 = ring[i]

        if p1[1] <= p[1]:
            if p2[1] > p[1]:
                if ccw(p1, p2, p) > 0:
                    wn += 1
        else:
            if p2[1] <= p[1]:
                if ccw(p1, p2, p) < 0:
                    wn -= 1
    return wn


def p_in_polygon(p, polygon):
    '''Test, whether `p` is in the polygon.

    Parameters:
        p: Tuple[float, float]           2D Point.
        polygon: List[List[Tuple[float, float]]]  Polygon, i.e. one exterior ring,
                                                  and multiple interior rings (holes).

    Returns:
        bool: True, if `p` in `polygon`, False otherwise.
    '''
    assert len(polygon) >= 1

    if winding_number(p, polygon[0]) != 0:
        # p inside polygon
        # check p not inside one of the holes
        for hole in polygon[1:]:
            if winding_number(p, hole) != 0:
                return False  # in hole of polygon
        return True

    return False



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

# The MIT License
#
# Copyright (c) 2017 Tammo Ippen, tammo.ippen@posteo.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from geohash2 import encode

from ._geo_fkt import bbox_hash, in_bbox
from ._geojson import geojson

try:
    from ._shapely import prepare, p_in_polygon
except ImportError:
    from ._pure import prepare, p_in_polygon


_SHAPES = None
def shapes():  # noqa: E302
    '''Return preprocessed shapes for `search` lookup.

    Preprocess geojson features depending on the search backend. Group shapes by
    the geohash of its bounding box, i.e. the first geohash rectangle that fully
    contains the bbox. This results (more-or-less) in a quad-tree and gives fast
    lookups (best case O(log(n)); worst case still O(n) ).

    Returns:
        Dict[str, List[Dict[str, Any]]]: Preprocessed shapes.
    '''
    global _SHAPES
    if _SHAPES is not None:
        return _SHAPES

    data = geojson()

    _SHAPES = dict()  # geohash -> shapes
    for feat in data['features']:
        for shp in prepare(feat):
            key = bbox_hash(shp['bounds'])
            shp['geohash'] = key
            if key not in _SHAPES:
                _SHAPES[key] = []
            _SHAPES[key].append(shp)

    return _SHAPES


def search_all(lng, lat):
    '''Reverse geocode lng/lat coordinate within the features from `shapes`.

    Look within the features from the `shapes()` function for all polygon that
    contains the point (lng, lat). From all found feature the `porperties`
    will be returned (more or less sorted from smallest to largest feature).
    `None`, if no feature containes the point.

    Parameters:
        lng: float  Longitude (-180, 180) of point. (WGS84)
        lat: float  Latitude (-90, 90) of point. (WGS84)

    Returns:
        Iterator[Dict[Any, Any]]  Iterator for `properties` of found features.
    '''
    if not (-180 <= lng <= 180):
        raise ValueError('Longitude must be between -180 and 180.')
    if not (-90 <= lat <= 90):
        raise ValueError('Latitude must be between -90 and 90.')

    key = encode(latitude=lat, longitude=lng)
    for sub_key in [key] + [key[:-i] for i in range(1, len(key) + 1)]:
        # look withing geohash rectangles of increasing resolution
        for shp in shapes().get(sub_key, []):
            # look through all shapes within one resolution
            if in_bbox((lng, lat), shp['bounds']):
                # first check if point in bbox
                if p_in_polygon((lng, lat), shp):
                    # ensure point is in polygon
                    yield shp['properties']
                    # look for other overlaps


def search(lng, lat):
    '''Reverse geocode lng/lat coordinate within the features from `shapes`.

    Look within the features from the `shapes()` function for a polygon that
    contains the point (lng, lat). From the first found feature the `porperties`
    will be returned. `None`, if no feature containes the point.

    Parameters:
        lng: float  Longitude (-180, 180) of point. (WGS84)
        lat: float  Latitude (-90, 90) of point. (WGS84)

    Returns:
        Dict[Any, Any]  `Properties` of found feature. `None` if nothing is found.
    '''
    try:
        return next(search_all(lng, lat))
    except StopIteration:
        return None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

# The MIT License
#
# Copyright (c) 2017 Tammo Ippen, tammo.ippen@posteo.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from ._geo_fkt import bbox
from ._geo_fkt import p_in_polygon as pure_p_in_polygon


def prepare(feat):
    '''Prepare geojson feature for further processing in `geopip._pure.p_in_polygon()`

    Parameters:
        feat: Dict[str, Any]  Geojson feature (only Polygon and MultiPolygon will
                              be processed).

    Returns:
        List[Dict[str, Any]]  Prepared shapes for `geopip._pure.p_in_polygon()`
    '''
    shp = feat['geometry']
    res = []
    if shp['type'] == 'MultiPolygon':
        for p_coords in shp['coordinates']:
            polygon = dict(
                type='Polygon',
                coordinates=p_coords
            )
            res += [dict(
                shape=polygon,
                properties=feat['properties'],
                bounds=bbox(polygon)
            )]
    elif shp['type'] == 'Polygon':
        res += [dict(
            shape=shp,
            properties=feat['properties'],
            bounds=bbox(shp)
        )]
    return res


def p_in_polygon(p, shp):
    '''Test, whether point `p` is in shape `shp`.

    Use the pure python implementation.

    Parameters:
        p:   Tuple[float, float]  Point (lng, lat) in WGS84.
        shp: Dict[str, Any]       Prepared shape dictionary from `geopip._pure.prepare()`.

    Returns:
        boolean: True, if p in shp, False otherwise
    '''
    return pure_p_in_polygon(p, shp['shape']['coordinates'])



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

# The MIT License
#
# Copyright (c) 2017 Tammo Ippen, tammo.ippen@posteo.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import json
from os import environ

import pkg_resources


def geojson():
    '''Load geojson file.

    If environment variable `REVERSE_GEOCODE_DATA` is set, load that file, otherwise
    load provided `globe.geo.json` file.

    Returns:
        Dict[Any, Any]: Geojson decoded as python dict.
    '''
    if environ.get('REVERSE_GEOCODE_DATA'):
        with open(environ['REVERSE_GEOCODE_DATA'], 'r') as f:
            return json.load(f)
    return json.loads(pkg_resources.resource_string('geopip', 'globe.geo.json').decode())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from setuptools import find_packages, setup

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    long_description = open('README.md').read()


version = '0.3'

setup(
    name='geopip',
    version=version,
    packages=find_packages(),
    install_requires=[
        'geohash2',
        # 'numpy',
        # 'shapely[vectorized]>=1.6b4',
    ],
    author='Tammo Ippen',
    author_email='tammo.ippen@posteo.de',
    description='Reverse geocode a lng/lat coordinate within a geojson FeatureCollection.',
    long_description=long_description,
    url='https://github.com/tammoippen/geopip',
    download_url='https://github.com/tammoippen/geopip/archive/v{}.tar.gz'.format(version),
    keywords=['geojson', 'point in polygon', 'reverse geocode', 'countries'],
    include_package_data=True,
)
